/*
  MeCab -- Yet Another Part-of-Speech and Morphological Analyzer

  $Id: feature_index.cpp,v 1.9 2006/02/28 14:52:48 taku-ku Exp $;

  Copyright (C) 2001-2005 Taku Kudo <taku@chasen.org>
  Copyright (C) 2004-2005 Nippon Telegraph and Telephone Corporation

  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <fstream>
#include <string>
#include <cstring>

#include "common.h"
#include "strutil.h"
#include "learner_node.h"
#include "feature_index.h"
#include "string_buffer.h"

#define BUFSIZE (2048)
#define POSSIZE (64)

#define ADDB(b) do { int id = this->id((b)); \
   if (id != -1) feature_.push_back (id); } while(0)

#define COPY_FEATURE(ptr) do {  \
   feature_.push_back (-1); \
   (ptr) = feature_freelist_.alloc (feature_.size());  \
   std::copy (feature_.begin(), feature_.end(), (ptr)); \
   feature_.clear (); } while(0)

namespace MeCab {

  namespace {
    char getEscapedChar (const char p)
    {
      switch (p) {
      case '0':  return '\0'; break;
      case 'a':  return '\a'; break;
      case 'b':  return '\b'; break;
      case 't':  return '\t'; break;
      case 'n':  return '\n'; break;
      case 'v':  return '\v'; break;
      case 'f':  return '\f'; break;
      case 'r':  return '\r'; break;
      case 's':  return ' ';  break; // space
      case '\\': return '\\'; break;
      default:  return p; break;
      }

      return '\0'; // never be here
    }
  }

  char* FeatureIndex::getIndex (char* &p, char ** column, size_t max)
  {
    ++p;

    bool flg = false;

    if (*p == '?') { flg = true; ++p; } // undef flg

    if (*p !='[') {
      WHAT("getIndex(): unmatched '['");
      return 0;
    }

    size_t n = 0;
    ++p;

    for (;; ++p) {
      switch (*p) {
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
	n = 10 * n + (*p - '0');
	break;
      case ']':
	if (n >= max) {
	  WHAT("given index is out of range: " << n);
	  return 0;
	}

	if (flg == true &&
	    (std::strcmp ("*", column[n]) == 0)
	    || column[n][0] == '\0') return 0;

	return column[n]; // return;

	break;
      default:
	WHAT("unmatched '['");
	return 0;
      }
    }

    return 0;
  }

  void FeatureIndex::set_alpha (double *alpha)
  {
    alpha_ = alpha;
  }

  bool FeatureIndex::openTemplate (Param &param)
  {
    std::string filename = createFileName
      (param.getProfileString ("dicdir"), FEATURE_DEF_FILE);

    std::ifstream ifs (filename.c_str());
    if (! ifs)
      WHAT_FALSE("no such file or directory: " << filename);

    char buf[8192];
    char *column[4];

    unigram_templs_.clear();
    bigram_templs_.clear();

    while (ifs.getline (buf, sizeof(buf))) {
      if (buf[0] == '\0' || buf[0] == '#' || buf[0] == ' ') continue;
      if (tokenize (buf, "\t ", column, 2) != 2)
	WHAT_FALSE("format error: " <<filename);

      if ( strcmp (column[0], "UNIGRAM") == 0 )
	unigram_templs_.push_back (this->strdup(column[1]));
      else if ( strcmp (column[0], "BIGRAM") == 0 )
	bigram_templs_.push_back (this->strdup(column[1]));
      else
	WHAT_FALSE("format error: " <<  filename);

    }

    return true;
  }

  bool EncoderFeatureIndex::open (Param &param)
  {
    return openTemplate(param);
  }

  bool DecoderFeatureIndex::open (Param &param)
  {
    std::string modelfile = param.getProfileString ("model");

    if (! mmap_.open (modelfile.c_str()))
      WHAT_CLOSE_FALSE(mmap_.what());

    char *ptr = mmap_.begin ();
    unsigned int maxid;
    read_static<unsigned int>(&ptr, maxid);
    maxid_ = static_cast<size_t>(maxid);
    alpha_ = (double *)ptr;
    ptr += (sizeof (double) * maxid_);
    da_.set_array ((void *)ptr);

    if (! openTemplate (param)) {
      close();
      return false;
    }

    return true;
  }

  void DecoderFeatureIndex::clear ()
  {
    feature_freelist_.free ();
  }

  void EncoderFeatureIndex::clear () {}

  void EncoderFeatureIndex::clearcache ()
  {
    feature_cache_.clear();
  }

  void EncoderFeatureIndex::close ()
  {
    dic_.clear ();
    feature_cache_.clear ();
    maxid_ = 0;
  }

  void DecoderFeatureIndex::close ()
  {
    da_.clear ();
    mmap_.close ();
    maxid_ = 0;
  }

  void FeatureIndex::calcCost (LearnerNode *node)
  {
    node->wcost = 0.0;
    if (node->stat == MECAB_EOS_NODE) return;
    for (int *f = node->fvector; *f != -1; ++f) node->wcost += alpha_[*f];
  }

  void FeatureIndex::calcCost (LearnerPath *path)
  {
    if ( is_empty (path) ) return;
    path->cost = path->rnode->wcost; // initialize;
    for (int *f = path->fvector; *f != -1; ++f) path->cost += alpha_[*f];
  }

  char *FeatureIndex::strdup (const char *p)
  {
    size_t len = std::strlen (p);
    char *q = char_freelist_.alloc (len+1);
    std::strcpy (q, p);
    return q;
  }

  bool DecoderFeatureIndex::buildFeature (LearnerPath *path)
  {
    path->rnode->wcost = path->cost = 0.0;
    if (! buildUnigramFeature (path)) return false;
    if (! buildBigramFeature (path))  return false;
    return true;
  }

  bool EncoderFeatureIndex::buildFeature (LearnerPath *path)
  {
    path->rnode->wcost = path->cost = 0.0;

    {
      std::map<std::string, int *>::iterator it = feature_cache_.find (path->rnode->ufeature);
      if (it != feature_cache_.end()) {
	path->rnode->fvector = it->second;
      } else {
	if (! buildUnigramFeature (path)) return false;
	feature_cache_.insert (std::pair<std::string, int *>(path->rnode->ufeature, path->rnode->fvector));
      }
    }

    {
      os_.clear();
      os_ << path->lnode->rfeature << ' ' << path->rnode->lfeature << '\0';
      std::string key(os_.str());
      std::map<std::string, int *>::iterator it = feature_cache_.find (key);
      if (it != feature_cache_.end()) {
	path->fvector = it->second;
      } else {
	if (! buildBigramFeature (path)) return false;
	feature_cache_.insert (std::pair<std::string, int *>(key, path->fvector));
      }
    }

    return true;
  }


  bool FeatureIndex::buildUnigramFeature (LearnerPath *path)
  {
    char ubuf[BUFSIZE];
    char *F[POSSIZE];

    feature_.clear ();
    strncpy (ubuf, path->rnode->ufeature, BUFSIZE);
    size_t usize = tokenizeCSV (ubuf, F, POSSIZE);

    for (std::vector<char*>::iterator it = unigram_templs_.begin();
	 it != unigram_templs_.end(); ++it) {
      char *p = *it;
      os_.clear();

      for (; *p; p++) {
	switch (*p) {
	default: os_ << *p; break;
	case '\\': os_ << getEscapedChar (*++p); break;
	case '%': {
	  switch (*++p) {
	  case 'F':  {
	    char *r = getIndex (p, F, usize); if (! r) goto NEXT;
	    os_ << r;
	  } break;
	  case 't':  os_ << (size_t)path->rnode->char_type;     break;
	  default:
	    WHAT_FALSE("unkonwn meta char: " <<  *p);
	  }
	}
	}
      }

      os_ << '\0';
      ADDB(os_.str());

    NEXT: continue;
    }

    COPY_FEATURE(path->rnode->fvector);

    return true;
  }

  bool FeatureIndex::buildBigramFeature (LearnerPath *path)
  {
    char rbuf[BUFSIZE];
    char lbuf[BUFSIZE];
    char *R[POSSIZE];
    char *L[POSSIZE];

    feature_.clear ();
    strncpy  (lbuf,  path->lnode->rfeature, BUFSIZE);
    strncpy  (rbuf,  path->rnode->lfeature, BUFSIZE);

    size_t lsize = tokenizeCSV (lbuf, L, POSSIZE);
    size_t rsize = tokenizeCSV (rbuf, R, POSSIZE);

    for (std::vector<char*>::iterator it = bigram_templs_.begin();
	 it != bigram_templs_.end(); ++it) {
      char *p = *it;
      os_.clear();

      for (; *p; p++) {
	switch (*p) {
	default: os_ << *p; break;
	case '\\': os_ << getEscapedChar (*++p); break;
	case '%': {
	  switch (*++p) {
	  case 'L': {
	    char *r = getIndex (p, L, lsize); if (! r) goto NEXT;
	    os_ << r;
	  } break;
	  case 'R': {
	    char *r = getIndex (p, R, rsize); if (! r) goto NEXT;
	    os_ << r;
	  } break;
	  default:
	    WHAT_FALSE("unkonwn meta char: " <<  *p);
	  }
	}
	}
      }

      os_ << '\0';

      ADDB(os_.str());

    NEXT: continue;
    }

    COPY_FEATURE(path->fvector);

    return true;
  }

  int DecoderFeatureIndex::id (const char *key)
  {
    return da_.exactMatchSearch<Darts::DoubleArray::result_type>(key);
  }

  int EncoderFeatureIndex::id (const char *key)
  {
    std::map <std::string, std::pair<int, unsigned int> >::iterator it = dic_.find(key);
    if (it == dic_.end()) {
      dic_.insert (std::make_pair<std::string, std::pair<int, unsigned int> >
		   (std::string (key), std::make_pair<int, unsigned int>((int)maxid_, 1)));
      int n = (int)maxid_;
      ++maxid_;
      return n;
    } else {
      it->second.second++;
      return it->second.first;
    }
    return -1;
  }

  void EncoderFeatureIndex::shrink (size_t freq,
				    std::vector<double> &observed)
  {
    if (freq <= 1) return;

    std::map<int, int> old2new;
    int new_maxid = 0;

    for (std::map<std::string, std::pair<int, unsigned int> >::iterator
	   it = dic_.begin(); it != dic_.end();) {
      if (it->second.second >= freq) {
	old2new.insert(std::make_pair<int,int>(it->second.first, new_maxid));
	it->second.first = new_maxid;
	++new_maxid;
	++it;
      } else {
	dic_.erase (it++);
      }
    }

    maxid_ = new_maxid;

    // replace all fvector
    for (std::map<std::string, int*>::iterator
	   it = feature_cache_.begin(); it != feature_cache_.end(); ++it) {
      int *to = it->second;
      for (int *f = it->second; *f != -1; ++f) {
	std::map<int,int>::iterator it2 = old2new.find(*f);
	if (it2 != old2new.end()) {
	  *to = it2->second;
	  ++to;
	}
      }
      *to = -1;
    }

    // replace observed vector
    std::vector<double> observed_new(maxid_);
    for (size_t i = 0; i < observed.size(); ++i) {
      std::map<int,int>::iterator it = old2new.find((int)i);
      if (it != old2new.end())
	observed_new[it->second] = observed[i];
    }

    observed = observed_new; // copy

    return;
  }

  bool EncoderFeatureIndex::convert (const char* txtfile, const char *binfile)
  {
    std::ifstream ifs (txtfile);

    if (!ifs )
      WHAT_FALSE("no such file or directory: " << txtfile);

    char buf[8192];
    char *column[4];
    std::map<std::string, double> dic;

    while (ifs.getline (buf, sizeof(buf))) {
      if (tokenize (buf, "\t ", column, 2) != 2)
	WHAT_FALSE("format error: " << buf);

      dic.insert (std::make_pair<std::string, double>
		  (std::string(column[1]), atof(column[0]) ));
    }

    std::ofstream ofs (binfile, std::ios::out | std::ios::binary);
    if (! ofs )
      WHAT_FALSE("permission denied: " << binfile);

    std::vector <char *> key;
    unsigned int size = (unsigned int)dic.size ();
    ofs.write ((char *)(&size), sizeof (unsigned int));

    for (std::map<std::string, double>::iterator it = dic.begin(); it != dic.end(); ++it) {
      key.push_back ((char *)it->first.c_str());
      ofs.write ((char *)&it->second, sizeof (double));
    }

    Darts::DoubleArray da;
    if (da.build (key.size(), &key[0], 0, 0, 0) != 0)
      WHAT_FALSE("unkown error in building double array: " << binfile);

    ofs.write ((char *)da.array(), da.unit_size () * da.size());

    return true;
  }

  bool EncoderFeatureIndex::save (const char *filename)
  {
    std::ofstream ofs (filename);

    if (! ofs)
      WHAT_FALSE("permission denied: " << filename);

    ofs.setf(std::ios::fixed,std::ios::floatfield);
    ofs.precision (24);

    for (std::map<std::string, std::pair<int, unsigned int> >::iterator it = dic_.begin();
	 it != dic_.end(); ++it)
      ofs << alpha_[it->second.first] << " " << it->first << std::endl;

    return true;
  }
}
