/*
  MeCab -- Yet Another Part-of-Speech and Morphological Analyzer

  $Id: learner.cpp,v 1.5 2006/04/29 17:23:21 taku-ku Exp $;

  Copyright (C) 2001-2005 Taku Kudo <taku@chasen.org>
  Copyright (C) 2004-2005 Nippon Telegraph and Telephone Corporation

  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stream_wrapper.h"
#include "common.h"
#include "learner.h"
#include "lbfgs.h"
#include "thread.h"

namespace MeCab
{

#ifdef MECAB_USE_THREAD
  class learner_thread: public thread {
  public:
    unsigned short start_i;
    unsigned short thread_num;
    size_t size;
    size_t micro_p;
    size_t micro_r;
    size_t micro_c;
    size_t err;
    double f;
    EncoderLearnerTagger **x;
    std::vector<double> expected;
    void run () {
      micro_p = micro_r = micro_c = err = 0;
      f = 0.0;
      std::fill(expected.begin(), expected.end(), 0.0);
      for (size_t i = start_i; i < size; i += thread_num) {
	f += x[i]->gradient (&expected[0]);
	err += x[i]->eval (micro_c, micro_p, micro_r);
      }
    }
  };
#endif   

  bool Learner::run (Param &param, const char *ifile, const char *ofile)
  {
    close();

    std::cout.setf(std::ios::fixed, std::ios::floatfield);
    std::cout.precision(5);

    double C             = param.getProfileFloat ("cost");
    double eta           = param.getProfileFloat ("eta");
    bool   text_only     = param.getProfileFloat ("text-only");
    size_t eval_size     = param.getProfileInt   ("eval-size");
    size_t unk_eval_size = param.getProfileInt   ("unk-eval-size");
    size_t freq          = param.getProfileInt   ("freq");
    size_t thread_num    = param.getProfileInt   ("thread");

    if (C <= 0)
      WHAT_CLOSE_FALSE("cost parameter is out of range");

    if (eta <= 0)
      WHAT_CLOSE_FALSE("eta is out of range");

    if (eval_size <= 0)
      WHAT_CLOSE_FALSE("eval-size is out of range: " << eval_size);

    if (unk_eval_size <= 0)
      WHAT_CLOSE_FALSE("unk-eval-size is out of range: " << unk_eval_size);

    if (freq < 0)
      WHAT_CLOSE_FALSE("freq is out of range: " << unk_eval_size);
     
    if (thread_num > 1024 || thread_num <= 0)
      WHAT_CLOSE_FALSE("# thread is invalid: " << thread_num);

    if (! tokenizer_.open (param))
      WHAT_CLOSE_FALSE(tokenizer_.what ());

    if (! feature_index_.open (param))
      WHAT_CLOSE_FALSE(feature_index_.what ());

    istream_wrapper ifs (ifile);
    if (! *ifs)
      WHAT_CLOSE_FALSE("no such file or directory: " << ifile);

    std::cout << "reading corpus ..." << std::flush;

    while (*ifs) {
      EncoderLearnerTagger *_x = new EncoderLearnerTagger ();

      if (! _x->open (&tokenizer_, &path_freelist_,
		      &feature_index_,
		      eval_size,
		      unk_eval_size))
	WHAT_CLOSE_FALSE(_x->what());

      if (! _x->read (*ifs, observed_))
	WHAT_CLOSE_FALSE(_x->what());

      if (! _x->empty()) x_.push_back (_x);
      else delete _x;

      if (x_.size() % 100 == 0)
	std::cout << x_.size() << "... " << std::flush;
    }

    // shrink vector
    feature_index_.shrink(freq, observed_);
    feature_index_.clearcache ();

    int converge = 0;
    double old_f = 0.0;
    size_t psize = feature_index_.size();
    observed_.resize (psize);
    LBFGS lbfgs;

    alpha_.resize (psize);
    expected_.resize (psize);
    std::fill (alpha_.begin(), alpha_.end(), 0.0);
    lbfgs.init ((int)psize, 5);

    feature_index_.set_alpha (&alpha_[0]);

    std::cout << std::endl;
    std::cout << "Number of sentences: " << x_.size() << std::endl;
    std::cout << "Number of features:  " << psize    << std::endl;
    std::cout << "eta:                 " << eta      << std::endl;
    std::cout << "freq:                " << freq     << std::endl;
    std::cout << "C(sigma^2):          " << C        << std::endl << std::endl;

#ifdef MECAB_USE_THREAD
    std::vector<learner_thread> thread(thread_num);
    for (size_t i = 0; i < thread_num; ++i) {
      thread[i].start_i = i;
      thread[i].size = x_.size();
      thread[i].thread_num = thread_num;
      thread[i].x = &x_[0];
      thread[i].expected.resize(expected_.size());
    }
#endif

    for (size_t itr = 0; ;  ++itr) {

      std::fill (expected_.begin(), expected_.end(), 0.0);

      double f = 0.0;
      size_t err = 0;
      size_t micro_p = 0;
      size_t micro_r = 0;
      size_t micro_c = 0;

#ifdef MECAB_USE_THREAD
      for (size_t i = 0; i < thread_num; ++i) 
        thread[i].start();
      
      for (size_t i = 0; i < thread_num; ++i) 
        thread[i].join();

      for (size_t i = 0; i < thread_num; ++i) {
        f += thread[i].f;
        err += thread[i].err;
        micro_r += thread[i].micro_r;
        micro_p += thread[i].micro_p;
        micro_c += thread[i].micro_c;
        for (size_t k = 0; k < psize; ++k) 
          expected_[k] += thread[i].expected[k];
      }
#else
      for (size_t i = 0; i < x_.size(); ++i) {
	f += x_[i]->gradient (&expected_[0]);
	err += x_[i]->eval (micro_c, micro_p, micro_r);
      }
#endif

      double p = 1.0 * micro_c / micro_p;
      double r = 1.0 * micro_c / micro_r;
      double micro_f = 2 * p * r / (p + r);

      for (size_t i = 0; i < psize; ++i) {
	f += (alpha_[i] * alpha_[i]/(2.0 * C));
	expected_[i] = expected_[i] - observed_[i] + alpha_[i]/C;
      }

      double diff = (itr == 0 ? 1.0 : std::fabs(1.0 * (old_f - f) )/old_f);
      std::cout << "iter="    << itr
		<< " err="    << 1.0 * err/x_.size()
		<< " F="      << micro_f
		<< " target=" << f
		<< " diff="   << diff << std::endl;
      old_f = f;

      if (diff < eta) converge++; else converge = 0;
      if (converge == 3)  break; // 3 is ad-hoc

      int ret = lbfgs.optimize (&alpha_[0], &f, &expected_[0]);

      if (ret <= 0) WHAT_CLOSE_FALSE(lbfgs.what());
    }

    std::cout << "\nDone! writing model file ... " << std::endl;

    std::string txtfile = std::string (ofile);
    txtfile += ".txt";

    if (! feature_index_.save (txtfile.c_str()))
      WHAT_CLOSE_FALSE(feature_index_.what ());

    if (! text_only) {
      if (! feature_index_.convert (txtfile.c_str(), ofile))
	WHAT_CLOSE_FALSE(feature_index_.what ());
    }

    return true;
  }

  void Learner::close()
  {
    for (size_t i = 0; i < x_.size(); ++i)
      delete x_[i];
    x_.clear ();
    feature_index_.clear ();
    tokenizer_.clear ();
    path_freelist_.free ();
    expected_.clear ();
    observed_.clear ();
    alpha_.clear ();
  }
}
