/*
  MeCab -- Yet Another Part-of-Speech and Morphological Analyzer

  $Id: strutil.h,v 1.5 2006/03/07 11:47:21 taku-ku Exp $;

  Copyright (C) 2001-2005 Taku Kudo <taku@chasen.org>
  Copyright (C) 2004-2005 Nippon Telegraph and Telephone Corporation

  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef _MECAB_STRUTIL_H
#define _MECAB_STRUTIL_H

#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <string>
#include <algorithm>

namespace MeCab {

  void inline dtoa (double val, char *s)
  {
    std::sprintf(s, "%-16f", val);
    char *p = s;
    for (; *p != ' '; ++p) {};
    *p = '\0';
    return;
  }

  template <class T>
  inline void itoa (T val, char *s)
  {
    char *t;
    T mod;

    if(val < 0) {
      *s++ = '-';
      val = -val;
    }
    t = s;

    while (val) {
      mod = val % 10;
      *t++ = (char)mod + '0';
      val /= 10;
    }

    if (s == t) *t++ = '0';
    *t = '\0';
    std::reverse (s, t);

    return;
  }

  template <class T>
  inline void uitoa (T val, char *s)
  {
    char *t;
    T mod;
    t = s;

    while(val) {
      mod = val % 10;
      *t++ =(char)mod + '0';
      val /= 10;
    }

    if (s == t) *t++ = '0';
    *t = '\0';
    std::reverse (s, t);

    return;
  }

  inline char *read_ptr (char **ptr, size_t size)
  {
    char *r = *ptr;
    *ptr += size;
    return r;
  }

  template <class T>
  inline void read_static (char **ptr, T& value)
  {
    char *r = read_ptr (ptr, sizeof (T));
    memcpy (&value, r, sizeof (T));
  }

  inline char *mystrdup (const char *str)
  {
    size_t l = strlen (str);
    char *r = new char [l + 1];
    std::strncpy (r, str, l+1);
    return r;
  }

  inline char *mystrdup (const std::string &str)
  {
    return mystrdup(str.c_str());
  }

  inline std::string createFileName (const std::string &path,
				     const std::string &file)
  {
    std::string s = path;
#if defined (_WIN32) && !defined (__CYGWIN__)
    if (s.size () && s[s.size()-1] != '\\') s += '\\';
#else
    if (s.size () && s[s.size()-1] != '/') s += '/';
#endif
    s += file;
    return s;
  }

  inline void removeFileName (std::string &s)
  {
    int len = (int)s.size() - 1;
    bool ok = false;
    for (; len >= 0; --len) {
#if defined (_WIN32) && !defined (__CYGWIN__)
      if (s[len] == '\\') { ok = true; break; }
#else
      if (s[len] == '/')  { ok = true; break; }
#endif
    }
    if (ok) s = s.substr (0, len);
    else s = ".";
  }

  inline void replaceString (std::string &s,
			     const std::string &src,
			     const std::string &dst)
  {
    size_t pos = s.find (src);
    if (pos != std::string::npos) s.replace (pos, src.size(), dst);
  }
  template <class Iterator>
  inline size_t tokenizeCSV (char *str, Iterator out, size_t max)
  {
    char *stre = str + strlen (str);
    size_t size = 0;

    while (str < stre && size < max) {
      if (*str == '"') {
	char *p = 0;
	bool quoted = false; 
	for (p = str + 1; p < stre; ++p) {
	  if (*p == '"') {
	    if (*(p + 1) == ',' || *(p + 1) == '\0') {
	      *p = '\0';
	      break;
	    } else if (*(p + 1) == '"') {
	      quoted = true;
	      ++p;
	    } else {
	       // do nothing
	    }
	  } 
	}
	if (quoted)
	   for (char *q = str; *q != '\0'; ++q)
	     if (*q == '"' && *(q + 1) == '"')
	       for (char *r = (q + 1); *r != '\0'; ++r) 
		 *r = *(r + 1);
	quoted = false; 
	++size;
	*out++ = str + 1;
	str = p + 2;
      } else {
	char *p = std::find(str, stre, ',');
	*p = '\0';
	++size;
	*out++ = str;
	str = p + 1;
      }
    }

    return size;
  }

  template <class Iterator>
  inline size_t tokenize (char *str, const char *del,
			  Iterator out, size_t max)
  {
    char *stre = str + strlen (str);
    const char *dele = del + strlen (del);
    size_t size = 0;

    while (size < max) {
      char *n = std::find_first_of (str, stre, del, dele);
      *n = '\0';
      *out++ = str;
      ++size;
      if (n == stre) break;
      str = n + 1;
    }

    return size;
  }
}

#endif
