/*
  MeCab -- Yet Another Part-of-Speech and Morphological Analyzer

  $Id: tokenizer.cpp,v 1.12 2006/02/07 20:08:48 taku-ku Exp $;

  Copyright (C) 2001-2005 Taku Kudo <taku@chasen.org>
  Copyright (C) 2004-2005 Nippon Telegraph and Telephone Corporation

  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "tokenizer.h"
#include "common.h"
#include "strutil.h"
#include "darts.h"
#include "scoped_ptr.h"

#define DRESULT_SIZE (512)

namespace MeCab {

  namespace {

    void read_node_info (Dictionary &dic,
			 Token *token,
			 LearnerNode *&node)
    {
      node->lcAttr   = token->lcAttr;
      node->rcAttr   = token->rcAttr;
      node->posid    = token->posid;
      node->wcost2   = token->wcost;
      node->feature  = dic.feature(*token);
      node->ufeature = std::find(node->feature,  node->feature  + 512, '\0') + 1;
      node->lfeature = std::find(node->ufeature, node->ufeature + 512, '\0') + 1;
      node->rfeature = std::find(node->lfeature, node->lfeature + 512, '\0') + 1;
    }

    void read_node_info (Dictionary &dic,
			 Token *token,
			 Node *&node)
    {
      node->lcAttr  = token->lcAttr;
      node->rcAttr  = token->rcAttr;
      node->posid   = token->posid;
      node->wcost   = token->wcost;
      node->feature = dic.feature(*token);
    }

    void init_bos (Node &node) {}

    void init_bos (LearnerNode &node)
    {
      node.ufeature = node.feature;
      node.lfeature = node.feature;
      node.rfeature = node.feature;
    }
  }

  template class TokenizerImpl<Node, Path>;
  template class TokenizerImpl<LearnerNode, LearnerPath>;

#if __GNUC__ >= 4
  template TokenizerImpl<Node, Path>::TokenizerImpl();
  template void TokenizerImpl<Node, Path>::clear();
  template void TokenizerImpl<Node, Path>::close();
  template Node* TokenizerImpl<Node, Path>::getBOSNode();
  template Node* TokenizerImpl<Node, Path>::getEOSNode();
  template Node* TokenizerImpl<Node, Path>::lookup(const char*, const char*);
  template bool TokenizerImpl<Node, Path>::open(Param &);
  template TokenizerImpl<LearnerNode, LearnerPath>::TokenizerImpl();
  template void TokenizerImpl<LearnerNode, LearnerPath>::clear();
  template void TokenizerImpl<LearnerNode, LearnerPath>::close();
  template LearnerNode* TokenizerImpl<LearnerNode, LearnerPath>::getEOSNode();
  template LearnerNode* TokenizerImpl<LearnerNode, LearnerPath>::getBOSNode();
  template LearnerNode* TokenizerImpl<LearnerNode, LearnerPath>::lookup(const char*, const char*);
  template bool TokenizerImpl<LearnerNode, LearnerPath>::open(Param &);
#endif

  template <typename N, typename P>
  TokenizerImpl<N, P>::TokenizerImpl ():
    node_freelist_ (NODE_FREELIST_SIZE),
    daresults_(new Dictionary::result_type [DRESULT_SIZE]), id_(0) {}

  template <typename N, typename P>
  void TokenizerImpl<N, P>::clear ()
  {
    node_freelist_.free();
    id_ = 0;
  }

  template <typename N, typename P>
  N *TokenizerImpl<N, P>::getBOSNode ()
  {
    N *bosNode = getNewNode();
    memset (bosNode, 0, sizeof(N));
    bosNode->surface = const_cast<char *>("BOS/EOS"); // dummy
    bosNode->feature = bos_feature_.get();
    bosNode->isbest  = 1;
    bosNode->id      = id_ - 1;
    bosNode->stat = MECAB_BOS_NODE;
    init_bos (*bosNode);
    return bosNode;
  }

  template <typename N, typename P>
  N *TokenizerImpl<N, P>::getEOSNode ()
  {
    N *eosNode = getBOSNode (); // same
    eosNode->stat = MECAB_EOS_NODE;
    return eosNode;
  }

  template <typename N, typename P>
  bool TokenizerImpl<N, P>::open (Param &param)
  {
    close ();

    std::string prefix = param.getProfileString ("dicdir");

    if (! unkdic_.open (createFileName (prefix, UNK_DIC_FILE).c_str()))
      WHAT_CLOSE_FALSE(unkdic_.what());

    if (! property_.open (param))
      WHAT_CLOSE_FALSE(property_.what ());

    Dictionary *sysdic = new Dictionary;

    if (! sysdic->open (createFileName (prefix, SYS_DIC_FILE).c_str()))
      WHAT_CLOSE_FALSE(sysdic->what());

    if (sysdic->type() != 0)
      WHAT_CLOSE_FALSE("not a system dictionary: " << prefix);

    property_.set_charset (sysdic->charset());
    dic_.push_back (sysdic);

    std::string userdic = param.getProfileString ("userdic");
    if (! userdic.empty ()) {
      char buf[1024];
      char *_dic[256];
      std::strncpy(buf, userdic.c_str(), sizeof(buf));
      size_t n = tokenize(buf, ":", _dic, sizeof(_dic));
      for (size_t i = 0; i < n; ++i) {
	Dictionary *d = new Dictionary;
	if (! d->open (_dic[i]))
	  WHAT_CLOSE_FALSE(d->what());
	if (d->type() != 1)
	  WHAT_CLOSE_FALSE("not a user dictionary: " << _dic[i]);
	if (! sysdic->isCompatible(*d))
	  WHAT_CLOSE_FALSE("incompatible dictionary: " << _dic[i]);

	dic_.push_back(d);
      }
    }

    unk_tokens_.clear();
    for (size_t i = 0; i < property_.size(); ++i) {
      const char *key = property_.name (i);
      Dictionary::result_type n = unkdic_.exactMatchSearch(key);
      if (n.value == -1) WHAT_CLOSE_FALSE("cannot find UNK category: " << key);
      Token *token = unkdic_.token(n);
      size_t size  = unkdic_.token_size(n);
      unk_tokens_.push_back(std::make_pair(token, size));
    }

    space_ = property_.getCharInfo (0x20); // ad-hoc

    bos_feature_.reset_string (param.getProfileString ("bos-feature"));

    if (*bos_feature_ == '\0')
      WHAT_CLOSE_FALSE("bos-feature is undefined in dicrc");

    return true;
  }

#define ADDUNKNWON do { \
  Token  *token = unk_tokens_[cinfo.default_type].first; \
  size_t size   = unk_tokens_[cinfo.default_type].second; \
  for (size_t k = 0; k < size; ++k) { \
      N *newNode = getNewNode (); \
      read_node_info (unkdic_, token + k, newNode); \
      newNode->id = id_ - 1; \
      newNode->char_type = cinfo.default_type; \
      newNode->surface   = begin2; \
      newNode->length    = begin3 - begin2;\
      newNode->rlength   = begin3 - begin;\
      newNode->bnext     = resultNode; \
      newNode->stat      = MECAB_UNK_NODE; \
      resultNode         = newNode; } } while (0)

  template <typename N, typename P>
  N *TokenizerImpl<N, P>::lookup (const char *begin, const char *end)
  {
    CharInfo cinfo;
    N    *resultNode = 0;
    size_t mblen = 0;

    end = static_cast<size_t>(end - begin) >= 65535 ? begin + 65535 : end;

    char *begin2 = property_.seekToOtherType (begin, end, space_, cinfo, mblen);
    for (std::vector<Dictionary *>::iterator it = dic_.begin(); it != dic_.end(); ++it) {

      size_t n = (*it)->commonPrefixSearch(begin2, static_cast<size_t>(end - begin2),
					   daresults_.get(), DRESULT_SIZE);

      for (size_t i = 0; i < n; ++i) {

	size_t size  = (*it)->token_size(daresults_[i]);
	Token *token = (*it)->token(daresults_[i]);
	for (size_t j = 0; j < size; ++j) {
	  N *newNode   = getNewNode ();
	  read_node_info (*(*it), token + j, newNode);
	  newNode->length    = daresults_[i].length;
	  newNode->rlength   = begin2 - begin + newNode->length;
	  newNode->surface   = begin2;
	  newNode->stat      = MECAB_NOR_NODE;
	  newNode->char_type = cinfo.default_type;
	  newNode->bnext     = resultNode;
	  resultNode         = newNode;
	}
      }
    }

    if (resultNode && ! cinfo.invoke)  return resultNode;

    char *begin3 = begin2 + mblen;
    char *group_begin3 = 0;

    if (cinfo.group) {
      char *tmp = begin3;
      begin3 = property_.seekToOtherType (begin3, end, cinfo);
      ADDUNKNWON;
      group_begin3 = begin3;       
      begin3 = tmp;
    }

    for (size_t i = 1; i <= cinfo.length; ++i) {
      if (begin3 == group_begin3) continue;
      ADDUNKNWON;
      if (! cinfo.isKindOf (property_.getCharInfo (begin3, end, mblen))) break;
      begin3 += mblen;
    }

    if (! resultNode) { ADDUNKNWON; }

    return resultNode;
  }

#undef ADDUNKNWON

  template <typename N, typename P>
  void TokenizerImpl<N, P>::close ()
  {
    for (std::vector<Dictionary *>::iterator it = dic_.begin();
	 it != dic_.end(); ++it)
      delete *it;
    dic_.clear();
    unk_tokens_.clear();
    property_.close ();
  }
}
