/*
  MeCab -- Yet Another Part-of-Speech and Morphological Analyzer

  $Id: tool.cpp,v 1.9 2006/04/25 17:16:21 taku-ku Exp $;

  Copyright (C) 2001-2005 Taku Kudo <taku@chasen.org>
  Copyright (C) 2004-2005 Nippon Telegraph and Telephone Corporation

  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <map>

#include "mecab.h"
#include "tool.h"
#include "learner.h"
#include "param.h"
#include "darts.h"
#include "common.h"
#include "dictionary.h"
#include "stream_wrapper.h"
#include "token.h"
#include "strutil.h"

/* for MSVC */
#undef max
#undef min

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace std;

int progress_bar (const char* message, size_t current, size_t total)
{
  static char bar[] = "###########################################";
  static int scale = sizeof(bar) - 1;
  static int prev = 0;

  int cur_percentage  = (int)(100.0 * current/total);
  int bar_len        = (int)(1.0   * current*scale/total);

  if (prev != cur_percentage) {
    printf("%s: %3d%% |%.*s%*s| ", message, cur_percentage, bar_len, bar, scale - bar_len, "");
    if (cur_percentage == 100)  printf("\n");
    else      printf("\r");
    fflush(stdout);
  }

  prev = cur_percentage;

  return 1;
}

int progress_bar_darts (size_t current, size_t total)
{
  return progress_bar ("emitting double-array", current, total);
}

int mecab_tool (int argc, char **argv)
{
  MeCab::Tool tool;
  if (tool.run (argc, argv)) {
    return EXIT_SUCCESS;
  } else {
    std::cerr << tool.what () << std::endl;
    return EXIT_FAILURE;
  }
}

namespace MeCab {

  short int toCost (double d, int n)
  {
    short max =   32767;
    short min = - 32767;
    return static_cast<short> ( _max ( _min (-n * d, (double)max), (double)min) );
  }

  bool Tool::learn ()
  {
    std::string model = param_.getProfileString ("model");
    std::string ifile = param_.getProfileString ("input");

    if (model.empty ())
      WHAT_FALSE("model file is not given, use --model option");

    if (ifile.empty())
      WHAT_FALSE("input is not given, use --input option");

    if (! loadDicRc()) return false;

    Learner leaner;
    if (! leaner.run (param_, ifile.c_str(), model.c_str()))
      WHAT_FALSE(leaner.what());

    return true;
  }

  bool Tool::dicinfo ()
  {
    std::string ofile  = param_.getProfileString ("output");
    std::string ifile  = param_.getProfileString ("input");

    if (ifile.empty())
      WHAT_FALSE("dictionary is not given, use --input option");

    if (ofile.empty())
      WHAT_FALSE("output file is not given, use --output option");

    ostream_wrapper ofs (ofile.c_str());
    if (! *ofs)
      WHAT_FALSE("permission denied: " << ofile);

    Dictionary dic;
    if (! dic.open (ifile))
      WHAT_FALSE(dic.what());

    *ofs << "File Name:\t"     << ifile << std::endl;
    *ofs << "Version:\t"  << dic.version() << std::endl;
    *ofs << "Charset:\t"  << dic.charset() << std::endl;
    *ofs << "Size:\t" << dic.size() << std::endl;
    *ofs << "Learn mode:\t"    << dic.learn() << std::endl;
    *ofs << "Type:\t"    << dic.type()   << std::endl;

    return true;
  }

  bool Tool::test ()
  {
    if (! loadDicRc()) return false;

    DecoderLearnerTagger tagger;
    if (! tagger.open (param_) )
      WHAT_FALSE(tagger.what());

    while (tagger.parse (std::cin, std::cout)) {};

    return true;
  }

  bool Tool::mkbindic ()
  {
    std::string ofile = param_.getProfileString ("output");
    std::string tfile = param_.getProfileString ("tokenoutput");
    std::string ifile = param_.getProfileString ("input");

    if (ofile.empty())
      WHAT_FALSE("output double-array file is not given, use --output option");

    if (ifile.empty())
      WHAT_FALSE("input is not given, use --input option");

    istream_wrapper ifs (ifile.c_str());
    if (! *ifs)
      WHAT_FALSE("no such file or directory: " << ifile);

    char buf[8192];
    char* column[8];
    std::vector<std::pair<std::string, Token*> > dic;

    while (ifs->getline (buf, sizeof(buf))) {
      if (tokenize (buf, "\t", column, 8) != 7)
	WHAT_FALSE("format error: " << buf);
      Token* token = new Token;
      token->lcAttr   = std::atoi(column[1]);
      token->rcAttr   = std::atoi(column[2]);
      token->posid    = std::atoi(column[3]);
      token->wcost    = std::atoi(column[4]);
      token->feature  = std::atoi(column[5]);
      token->compound = std::atoi(column[6]);
      dic.push_back(std::make_pair<std::string, Token*>
		    (std::string(column[0]), token));
    }

    if (dic.empty())
      WHAT_FALSE("dictionary is empty: " << ifile);

    std::sort(dic.begin(), dic.end());

    size_t bsize = 0;
    size_t idx = 0;
    std::string prev = "";
    std::vector <const char *> str;
    std::vector <Darts::DoubleArray::result_type> val;

    for (size_t i = 0; i < dic.size(); ++i) {
      if (i != 0 && prev != dic[i].first) {
	str.push_back(dic[idx].first.c_str());
	val.push_back(bsize + (idx << 8));
	bsize = 1;
	idx = i;
      } else {
	++bsize;
      }
      prev = dic[i].first;
    }
    str.push_back(dic[idx].first.c_str());
    val.push_back(bsize + (idx << 8));

    Darts::DoubleArray da;
    if (da.build (str.size(), (char **)&str[0], 0, &val[0], &progress_bar_darts) < 0)
      WHAT_FALSE("unkown error in building double-array");

    if (0 != da.save(ofile.c_str(), "wb"))
      WHAT_FALSE("permission denied: " << ofile);      

    std::ofstream ofs (tfile.c_str(), std::ios::binary|std::ios::out);
    if (! ofs) WHAT_FALSE("permission denied: " << tfile);      

    for (size_t i = 0; i < dic.size(); ++i) {
      ofs.write((char *)dic[i].second, sizeof(Token));
      delete dic[i].second;
    }
    ofs.close();

    return true;
  }

  bool Tool::convmodel ()
  {
    std::string ofile = param_.getProfileString ("output");
    std::string ifile = param_.getProfileString ("input");

    if (ifile.empty())
      WHAT_FALSE("input text model file is not given, use --input option");

    if (ofile.empty())
      WHAT_FALSE("output binary model file is not given, use --output option");

    EncoderFeatureIndex fi;
    if (! fi.convert (ifile.c_str(), ofile.c_str()))
      WHAT_FALSE(fi.what ());

    return true;
  }

  bool Tool::convmatrix ()
  {
    std::string ofile = param_.getProfileString ("output");
    std::string ifile = param_.getProfileString ("input");

    if (ifile.empty())
      WHAT_FALSE("input text model file is not given, use --input option");

    if (ofile.empty())
      WHAT_FALSE("output binary model file is not given, use --output option");

    istream_wrapper ifs (ifile.c_str());
    if (! *ifs)
      WHAT_FALSE("No such file or directory: " << ifile);

    char *column[4];
    char buf[8192];

    ifs->getline (buf, sizeof(buf));

    if (tokenize (buf, "\t ", column, 2) != 2)
      WHAT_FALSE("format error: " << buf);

    unsigned short lsize = atoi (column[0]);
    unsigned short rsize = atoi (column[1]);
    std::vector <short> matrix (lsize * rsize);
    std::fill (matrix.begin(), matrix.end(), 0);

    while (ifs->getline (buf, sizeof(buf))) {

      if (tokenize (buf, "\t ", column, 4) != 3)
	WHAT_FALSE("format error: " << buf);

      size_t l = atoi (column[0]);
      size_t r = atoi (column[1]);
      int    c = atoi (column[2]);

      if (l >= lsize || r >= rsize)
	WHAT_FALSE("index values are out of range");

      progress_bar ("emitting matrix      ", l+1, lsize);

      matrix[(l + lsize * r)] = (short)c;
    }

    std::ofstream ofs (ofile.c_str(), std::ios::binary|std::ios::out);
    if (! ofs)
      WHAT_FALSE("permission denied: " << ofile);

    ofs.write ((char *)&lsize, sizeof (unsigned short));
    ofs.write ((char *)&rsize, sizeof (unsigned short));
    ofs.write ((char *)&matrix[0], lsize * rsize * sizeof (short));

    return true;
  }

  bool Tool::mktextdic ()
  {
    if (! loadDicRc()) return false;

    bool matrix       = param_.getProfileInt    ("matrix");
    bool unk          = param_.getProfileInt    ("unknown");
    int factor        = param_.getProfileInt    ("factor");
    std::string ofile = param_.getProfileString ("output");
    std::string ifile = param_.getProfileString ("input");
    std::string charset = param_.getProfileString ("charset");

    DecoderFeatureIndex fi;

    if (! fi.open (param_))
      WHAT_FALSE(fi.what());

    CharProperty property;
    if (! property.open (param_) )
      WHAT_FALSE(property.what());

    if (factor <= 0)
      WHAT_FALSE("non-positive cost factor is given");

    if (ifile.empty())
      WHAT_FALSE("input text file is not given, use --input option");

    if (ofile.empty())
      WHAT_FALSE("output file is not given, use --output option");

    ostream_wrapper ofs (ofile.c_str());
    if (! *ofs)
      WHAT_FALSE("permission denied: " << ofile);

    istream_wrapper ifs (ifile.c_str());
    if (! *ifs)
      WHAT_FALSE("no such file or directory: " << ifile);

    LearnerPath path;
    LearnerNode rnode;
    LearnerNode lnode;

    rnode.stat  = lnode.stat = MECAB_NOR_NODE;
    rnode.rpath = &path;
    lnode.lpath = &path;
    path.lnode  = &lnode;
    path.rnode  = &rnode;
    char buf[8192];
    char *column[8];

    if (matrix) {

      bool first = true;

      while (ifs->getline (buf, sizeof(buf))) {

	if (tokenize (buf, "\t", column, 8) < 6)
	  WHAT_FALSE("format error: " << buf);

	if (first) {
	  *ofs << column[0] << ' ' << column[1] << std::endl; // write size of matrix
	  first = false;
	}

	path.lnode->rfeature = column[3];
	path.rnode->lfeature = column[5];
	path.rnode->wcost = 0;
	fi.buildBigramFeature (&path);
	fi.calcCost (&path);
	*ofs << column[2] << ' ' << column[4] << ' '
	     << toCost (path.cost, factor) << std::endl;
      }

    } else {
       
      property.set_charset(charset.c_str());

      while (ifs->getline (buf, sizeof(buf))) {

	if (tokenize (buf, "\t", column, 8) != 5)
	  WHAT_FALSE("format error: " << buf);

	if (unk) {
	  int c = property.id (column[0]);

	  if (c < 0)
	    WHAT_FALSE("unknown property [" << column[0] << "]");

	  path.rnode->char_type = (unsigned char)c;

	} else {
	  size_t len = std::strlen (column[0]);
	  size_t mblen;
	  CharInfo cinfo = property.getCharInfo (column[0], column[0] + len, mblen);
	  path.rnode->char_type = cinfo.default_type;
	}

	path.rnode->ufeature  = column[4];
	fi.buildUnigramFeature (&path);
	fi.calcCost (&rnode);
	*ofs << column[0] << ',' << column[1] << ',' << column[2] << ','
	     << toCost (rnode.wcost, factor) << ',' << column[3] << std::endl;
      }
    }

    return true;
  }

  bool Tool::loadDicRc ()
  {
    std::string dicrc = createFileName (param_.getProfileString ("dicdir"), "dicrc");
    if (! param_.load (dicrc.c_str()))
      WHAT_FALSE("no such file or directory: " << dicrc);
    return true;
  }

  bool Tool::run (int argc, char **argv)
  {
    static const MeCab::Option long_options[] =
      {
	{ "learn",  'l',  0,    0,    "set training mode" },
	{ "test",  't',  0,    0,    "set test mode" },
	{ "textdic",  'D',  0,    0,   "make text dictionary"},
	{ "bindic",  'a',  0,    0,    "make binary dictionary" },
	{ "matrix",  'M',  0,    0,    "make connection matrix" },
	{ "dicinfo",  'k',  0,    0,    "set dictionary info mode" },
	{ "convmodel",  'b',  0,    0,    "convert text model to binary model" },
	{ "convmatrix", 'A',  0,    0,    "convert text matrix to binary matrix" },
	{ "dicdir",  'd',  ".",   "DIR",    "set DIR as dicdir (default \".\" )" },
	{ "unknown",  'u',  0,    0,    "build unknown dictionary" },
	{ "text-only",  'y',  0,    0,    "output text model only" },
	{ "factor",  'f',  "800",   "INT",    "use INT as cost factor (default 800)" },
	{ "freq",  'F',  "1",   "INT",    "feature frequency threshold (default 1)" },
	{ "output",  'o',  0,   "FILE",  "output filename" },
	{ "input",  'i',  0,   "FILE",  "input  filename" },
	{ "tokenoutput", 'T', 0, "FILE", "output filename for tokens"},
	{ "model",  'm',  0,   "FILE",  "use FILE as model file" },
	{ "cost",  'C',  "1.0",   "FLOAT", "use FLOAT as hyperparam_eter of CRF (default 1.0)" },
	{ "charset", 'c', 0, "STR", "dictionary charset"},
	{ "eta",  'e',  "0.0001",   "FLOAT", "use FLOAT as termination criterion (default 0.0001)" },
	{ "thread",  'p',  "1",   "NUM", "number of threads (default 1)" }, 
	{ "version",  'v',  0,   0,    "show the version and exit"   },
	{ "help",  'h',  0,   0,    "show this help and exit."   },
	{ 0, 0, 0, 0 }
      };

    param_.clear ();

    if (! param_.open (argc, argv, long_options)) {
      std::cout << param_.what () << "\n\n" <<  COPYRIGHT
		<< "\ntry '--help' for more information." << std::endl;
      return false;
    }

    if (param_.getProfileInt ("help")) {
      std::cout << param_.help (long_options);
      return true;
    }

    if (param_.getProfileInt ("version")) {
      std::cout << param_.version (long_options);
      return true;
    }

    if (param_.getProfileInt ("matrix")) param_.setProfile ("textdic", 1, true);

    if (param_.getProfileInt ("convmatrix")) return convmatrix ();
    if (param_.getProfileInt ("convmodel"))  return convmodel  ();
    if (param_.getProfileInt ("textdic"))    return mktextdic  ();
    if (param_.getProfileInt ("bindic"))     return mkbindic   ();
    if (param_.getProfileInt ("learn"))      return learn      ();
    if (param_.getProfileInt ("test"))       return test       ();
    if (param_.getProfileInt ("dicinfo"))    return dicinfo    ();

    std::cout << param_.help (long_options);

    return true;
  }
}
